// $Id$

/*************************************************************************
 * XDAQ Components for Distributed Data Acquisition                      *
 * Copyright (C) 2000-2009, CERN.			                 *
 * All rights reserved.                                                  *
 * Authors: J. Gutleber and L. Orsini					 *
 *                                                                       *
 * For the licensing terms see LICENSE.		                         *
 * For the list of contributors see CREDITS.   			         *
 *************************************************************************/
 
#include "xdata/xdata.h"
#include "xdata/XMLDOM.h"

#include "xdata/Serializable.h"
#include "xdata/soap/Serializer.h"
#include "xdata/exception/Exception.h"
#include "xdata/XStr.h"
#include "xdata/Float.h"
#include "xdata/Integer.h"
#include "xdata/Boolean.h"
#include "xdata/String.h"

#include "xoap/MessageFactory.h"
#include "xoap/SOAPEnvelope.h"
#include "xoap/SOAPName.h"
#include "xoap/SOAPBody.h"
#include "xoap/SOAPBodyElement.h"
#include "xoap/DOMParser.h"
#include "xoap/domutils.h"

#include "xcept/tools.h"

/* The SOAP message to be deserialized:

<soap-env:Envelope 
	soap-env:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/" 
	xmlns:soap-env="http://schemas.xmlsoap.org/soap/envelope/">
	<soap-env:Header/>
	<soap-env:Body>
		<xdaq:data 
			xmlns:soapenc="http://schemas.xmlsoap.org/soap/encoding" 
			xmlns:xdaq="urn:xdaq-soap:3.0" 
			xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" 
			xsi:type="soapenc:Struct">
				<xdaq:myBoolean xsi:type="xsd:boolean">true</xdaq:myBoolean>
				<xdaq:myFloat xsi:type="xsd:float">3.141000</xdaq:myFloat>
				<xdaq:myInteger xsi:type="xsd:integer">1</xdaq:myInteger>
				<xdaq:myString xsi:type="xsd:string">Example text</xdaq:myString>
		</xdaq:data>
	</soap-env:Body>
</soap-env:Envelope>
*/

int main(int argc, char** argv)
{
	// Initialize XML platform in standalone main program only
	XMLPlatformUtils::Initialize();
	
	xoap::DOMParser* parser = xoap::DOMParser::get("ParseFromSOAP");
	
	std::stringstream data;
	
	data << "<soap-env:Envelope " << std::endl;
	data << "soap-env:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\" " << std::endl;
	data << "xmlns:soap-env=\"http://schemas.xmlsoap.org/soap/envelope/\">" << std::endl;
	data << "<soap-env:Header/>" << std::endl;
	data << "<soap-env:Body>" << std::endl;
	data << "<xdaq:data" << std::endl;
	data << "xmlns:soapenc=\"http://schemas.xmlsoap.org/soap/encoding\" " << std::endl;
	data << "xmlns:xdaq=\"urn:xdaq-soap:3.0\" " << std::endl;
	data << "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " << std::endl;
	data << "xsi:type=\"soapenc:Struct\">" << std::endl;
	data << "<xdaq:myBoolean xsi:type=\"xsd:boolean\">true</xdaq:myBoolean>" << std::endl;
	data << "<xdaq:myFloat xsi:type=\"xsd:float\">3.141000</xdaq:myFloat>" << std::endl;
	data << "<xdaq:myInteger xsi:type=\"xsd:integer\">1</xdaq:myInteger>" << std::endl;
	data << "<xdaq:myString xsi:type=\"xsd:string\">Example text</xdaq:myString>" << std::endl;
	data << "</xdaq:data>" << std::endl;
	data << "</soap-env:Body>" << std::endl;
	data << "</soap-env:Envelope>" << std::endl;
	
	DOMDocument* doc = 0;
	try
	{
		doc = parser->parse (data.str());
	} 
	catch (xoap::exception::Exception& e)
	{
		std::cout << xcept::stdformat_exception_history(e) << std::endl;
		return 1;
	}
	
	// Declare a SOAP serializer
	xdata::soap::Serializer serializer;
	
	xdata::InfoSpace * infospace = xdata::InfoSpace::get("myInfoSpace");
	
	xdata::Integer i;
	infospace->fireItemAvailable("myInteger", &i, 0);
	
	xdata::Float f;
	infospace->fireItemAvailable("myFloat", &f, 0);
	
	xdata::Boolean b;
	infospace->fireItemAvailable("myBoolean", &b, 0);
	
	xdata::String s;
	infospace->fireItemAvailable("myString", &s, 0);
	
	// Find the <xdaq:xdata> tag
	DOMNodeList* dataNode = doc->getElementsByTagNameNS(xoap::XStr("urn:xdaq-soap:3.0"), xoap::XStr("data"));

	// We expect only one data node, so take the item with index 0	
	try
	{
		serializer.import ( infospace, dataNode->item(0) );
	}
	catch (xdata::exception::Exception& xde)
	{
		std::cout << xcept::stdformat_exception_history(xde) << std::endl;
	}
	
	// Dump the read data values
	std::cout << "Read integer: " << dynamic_cast<xdata::Integer*>(infospace->find("myInteger"))->toString() << std::endl;
	std::cout << "Read float: " << dynamic_cast<xdata::Float*>(infospace->find("myFloat"))->toString() << std::endl;
	std::cout << "Read boolean: " << dynamic_cast<xdata::Boolean*>(infospace->find("myBoolean"))->toString() << std::endl;
	std::cout << "Read string: " << dynamic_cast<xdata::String*>(infospace->find("myString"))->toString() << std::endl;
	
	// Remove the parser and therefore release the DOM document
	xoap::DOMParser::remove("ParseFromSOAP");
	
	return 0;
}
